"""Base client for Radarr/Sonarr API interactions."""

from __future__ import annotations

import asyncio
import contextlib
import hashlib
import logging
import time
from typing import TYPE_CHECKING, Any, Protocol, Self, runtime_checkable

import httpx
from cachetools import TTLCache
from tenacity import (
    RetryCallState,
    retry,
    retry_base,
    stop_after_attempt,
    wait_exponential,
)

# Status codes that should trigger a retry
RETRYABLE_STATUS_CODES = frozenset({429, 500, 502, 503, 504})


class RetryableHTTPError(Exception):
    """Exception for HTTP errors that should trigger a retry.

    This exception wraps HTTPStatusError for cases where we want to retry
    (429, 5xx responses) and carries the retry delay from Retry-After header.
    """

    def __init__(
        self,
        message: str,
        response: httpx.Response,
        retry_after: float | None = None,
    ) -> None:
        """Initialize RetryableHTTPError.

        Args:
            message: Error message
            response: The HTTP response that triggered the error
            retry_after: Optional delay in seconds from Retry-After header
        """
        super().__init__(message)
        self.response = response
        self.retry_after = retry_after

    @property
    def status_code(self) -> int:
        """Get the HTTP status code."""
        return self.response.status_code


class RetryPredicate(retry_base):
    """Custom retry predicate that retries on connection errors and retryable HTTP errors."""

    def __call__(self, retry_state: RetryCallState) -> bool:
        """Check if the exception should trigger a retry.

        Args:
            retry_state: The current retry state

        Returns:
            True if the exception should trigger a retry
        """
        if retry_state.outcome is None:
            return False

        exception = retry_state.outcome.exception()
        if exception is None:
            return False

        # Retry on connection errors
        if isinstance(exception, (httpx.ConnectError, httpx.ConnectTimeout, httpx.ReadTimeout)):
            return True

        # Retry on our custom retryable HTTP error
        return isinstance(exception, RetryableHTTPError)


class RetryAfterWait(wait_exponential):
    """Custom wait strategy that respects Retry-After header when present."""

    def __call__(self, retry_state: RetryCallState) -> float:
        """Calculate the wait time, respecting Retry-After if available.

        Args:
            retry_state: The current retry state

        Returns:
            Wait time in seconds
        """
        # Check if the exception has a retry_after value
        if retry_state.outcome is not None:
            exception = retry_state.outcome.exception()
            if isinstance(exception, RetryableHTTPError) and exception.retry_after is not None:
                return exception.retry_after

        # Fall back to exponential backoff
        return super().__call__(retry_state)


if TYPE_CHECKING:
    from filtarr.models.common import Release, Tag

logger = logging.getLogger(__name__)


@runtime_checkable
class ReleaseProvider(Protocol):
    """Protocol for clients that can fetch releases for media items.

    This protocol defines the interface for clients that can search for and
    return releases from indexers. Both RadarrClient and SonarrClient implement
    this protocol, allowing them to be used polymorphically in release-checking
    operations.

    Example:
        async def check_releases(provider: ReleaseProvider, item_id: int) -> bool:
            releases = await provider.get_releases_for_item(item_id)
            return any(r.is_4k() for r in releases)
    """

    async def get_releases_for_item(self, item_id: int) -> list[Release]:
        """Fetch releases for a specific media item.

        Args:
            item_id: The ID of the media item (movie ID for Radarr,
                     series ID for Sonarr)

        Returns:
            List of Release models found by indexers
        """
        ...


@runtime_checkable
class TaggableClient(Protocol):
    """Protocol for clients that support tag operations.

    This protocol defines the interface for clients that can manage tags
    on media items. Both RadarrClient and SonarrClient implement this protocol.
    """

    async def get_tags(self) -> list[Tag]:
        """Fetch all tags.

        Returns:
            List of Tag models
        """
        ...

    async def create_tag(self, label: str) -> Tag:
        """Create a new tag.

        Args:
            label: The tag label

        Returns:
            The created Tag model
        """
        ...

    async def add_tag_to_item(self, item_id: int, tag_id: int) -> Any:
        """Add a tag to an item (movie or series).

        Args:
            item_id: The item ID
            tag_id: The tag ID to add

        Returns:
            The updated item model
        """
        ...

    async def remove_tag_from_item(self, item_id: int, tag_id: int) -> Any:
        """Remove a tag from an item (movie or series).

        Args:
            item_id: The item ID
            tag_id: The tag ID to remove

        Returns:
            The updated item model
        """
        ...


class BaseArrClient:
    """Base client with retry and caching for Radarr/Sonarr APIs.

    This base class provides:
    - HTTP client management with connection pooling
    - Automatic retry with exponential backoff for transient failures
    - Per-client TTL caching for GET requests
    - Context manager protocol for resource cleanup

    Subclasses should implement specific API methods using the provided
    `_get()` method for cached requests or `_get_uncached()` for fresh data.
    """

    def __init__(
        self,
        base_url: str,
        api_key: str,
        *,
        timeout: float = 120.0,
        cache_ttl: int = 300,
        max_retries: int = 3,
        max_connections: int = 20,
        max_keepalive_connections: int = 10,
    ) -> None:
        """Initialize the client.

        Args:
            base_url: The base URL of the arr instance (e.g., http://localhost:7878)
            api_key: The API key for authentication
            timeout: Request timeout in seconds (default 120.0)
            cache_ttl: Cache time-to-live in seconds (default 300)
            max_retries: Maximum number of retry attempts (default 3)
            max_connections: Maximum number of concurrent connections in the pool
                (default 20). This limits the total number of simultaneous
                connections to the server.
            max_keepalive_connections: Maximum number of idle keep-alive connections
                to maintain in the pool (default 10). Keep-alive connections are
                reused for subsequent requests to avoid connection overhead.
        """
        self.base_url = base_url.rstrip("/")
        self.api_key = api_key
        self.timeout = timeout
        self.cache_ttl = cache_ttl
        self.max_retries = max_retries
        self.max_connections = max_connections
        self.max_keepalive_connections = max_keepalive_connections

        self._client: httpx.AsyncClient | None = None
        self._cache: TTLCache[str, Any] = TTLCache(maxsize=1000, ttl=cache_ttl)
        self._cache_lock = asyncio.Lock()
        # Pending requests for stampede protection (thundering herd prevention)
        # Maps cache_key -> (Event, result_holder)
        # result_holder is a dict with 'data' or 'error' key once completed
        self._pending_requests: dict[str, tuple[asyncio.Event, dict[str, Any]]] = {}

    async def __aenter__(self) -> Self:
        """Enter async context manager."""
        limits = httpx.Limits(
            max_connections=self.max_connections,
            max_keepalive_connections=self.max_keepalive_connections,
        )
        self._client = httpx.AsyncClient(
            base_url=self.base_url,
            headers={"X-Api-Key": self.api_key},
            timeout=self.timeout,
            limits=limits,
        )
        return self

    async def __aexit__(self, *args: Any) -> None:
        """Exit async context manager."""
        if self._client:
            await self._client.aclose()
            self._client = None

    @property
    def client(self) -> httpx.AsyncClient:
        """Get the HTTP client, raising if not in context.

        Returns:
            The httpx async client

        Raises:
            RuntimeError: If called outside of async context manager
        """
        if self._client is None:
            raise RuntimeError("Client must be used within async context manager")
        return self._client

    def _make_cache_key(self, endpoint: str, params: dict[str, Any] | None) -> str:
        """Generate a cache key from endpoint and parameters.

        Args:
            endpoint: The API endpoint path
            params: Query parameters

        Returns:
            A unique cache key string
        """
        params_str = str(sorted((params or {}).items()))
        key_data = f"{endpoint}:{params_str}"
        return hashlib.sha256(key_data.encode()).hexdigest()[:16]

    async def _get(self, endpoint: str, params: dict[str, Any] | None = None) -> Any:
        """Make a cached GET request with stampede protection.

        Checks the cache first. If not found, makes the request and caches
        the result. Includes stampede protection to prevent thundering herd
        when multiple concurrent requests hit the same uncached key.

        Args:
            endpoint: The API endpoint path (e.g., "/api/v3/release")
            params: Optional query parameters

        Returns:
            The JSON response data

        Raises:
            httpx.HTTPStatusError: On HTTP errors (after retries exhausted)
        """
        cache_key = self._make_cache_key(endpoint, params)

        # Variables to track if we're the leader or a waiter
        is_leader = False
        event: asyncio.Event | None = None
        result_holder: dict[str, Any] | None = None

        async with self._cache_lock:
            # Check cache first
            if cache_key in self._cache:
                logger.debug("Cache hit for %s", endpoint)
                return self._cache[cache_key]

            # Check if there's already a pending request for this key
            if cache_key in self._pending_requests:
                # Another request is already fetching this data - we'll wait for it
                event, result_holder = self._pending_requests[cache_key]
                logger.debug("Waiting for pending request for %s", endpoint)
            else:
                # We'll be the "leader" - create the pending request entry
                is_leader = True
                event = asyncio.Event()
                result_holder = {}
                self._pending_requests[cache_key] = (event, result_holder)
                logger.debug("Cache miss for %s, fetching from API", endpoint)

        # If we're a waiter (not the leader), wait for the leader to complete
        if not is_leader:
            assert event is not None and result_holder is not None
            await event.wait()

            # Return the result or re-raise the error
            if "error" in result_holder:
                raise result_holder["error"]
            return result_holder["data"]

        # We're the leader - fetch the data
        assert event is not None and result_holder is not None
        try:
            data = await self._get_uncached(endpoint, params)

            # Store in cache and set result for waiters
            async with self._cache_lock:
                self._cache[cache_key] = data
                result_holder["data"] = data
                # Clean up pending request entry
                self._pending_requests.pop(cache_key, None)

            # Signal waiters that data is ready (outside lock to avoid deadlock)
            event.set()

            return data

        except Exception as e:
            # Store error for waiters
            async with self._cache_lock:
                result_holder["error"] = e
                # Clean up pending request entry
                self._pending_requests.pop(cache_key, None)

            # Signal waiters that we're done (with error)
            event.set()
            raise

    async def _get_uncached(self, endpoint: str, params: dict[str, Any] | None = None) -> Any:
        """Make a GET request without caching.

        This method includes retry logic for transient failures.

        Args:
            endpoint: The API endpoint path
            params: Optional query parameters

        Returns:
            The JSON response data

        Raises:
            httpx.HTTPStatusError: On HTTP errors (after retries exhausted)
        """
        return await self._request_with_retry("GET", endpoint, params=params)

    async def _request_with_retry(
        self,
        method: str,
        endpoint: str,
        params: dict[str, Any] | None = None,
        json: dict[str, Any] | None = None,
    ) -> Any:
        """Make an HTTP request with retry logic.

        Retries on:
        - Connection errors
        - Timeouts
        - 429 (Too Many Requests)
        - 5xx server errors

        Does NOT retry on:
        - 401 (Unauthorized) - fail fast
        - 404 (Not Found) - fail fast
        - Other 4xx client errors

        Args:
            method: HTTP method (GET, POST, etc.)
            endpoint: The API endpoint path
            params: Optional query parameters
            json: Optional JSON body

        Returns:
            The JSON response data

        Raises:
            httpx.HTTPStatusError: On non-retryable HTTP errors
            tenacity.RetryError: After all retries exhausted
        """
        # Track timing for performance diagnostics
        start_time = time.monotonic()

        @retry(
            stop=stop_after_attempt(self.max_retries),
            wait=RetryAfterWait(multiplier=1, min=1, max=10),
            retry=RetryPredicate(),
            before_sleep=self._log_retry,
            reraise=True,
        )
        async def _do_request() -> Any:
            response = await self.client.request(
                method,
                endpoint,
                params=params,
                json=json,
            )

            # Don't retry on 401/404 - fail immediately
            if response.status_code in (401, 404):
                response.raise_for_status()

            # Retry on 429 or 5xx - raise RetryableHTTPError
            if response.status_code in RETRYABLE_STATUS_CODES:
                logger.warning("Retryable HTTP error %d for %s", response.status_code, endpoint)

                # Parse Retry-After header if present (for 429 responses)
                retry_after: float | None = None
                if response.status_code == 429:
                    retry_after_header = response.headers.get("Retry-After")
                    if retry_after_header:
                        # Retry-After can also be an HTTP date, but we only parse numeric values
                        with contextlib.suppress(ValueError):
                            retry_after = float(retry_after_header)

                raise RetryableHTTPError(
                    f"Retryable HTTP error {response.status_code} for {endpoint}",
                    response=response,
                    retry_after=retry_after,
                )

            # Other errors - raise without retry
            response.raise_for_status()

            return response.json()

        try:
            result = await _do_request()
            elapsed = time.monotonic() - start_time
            # Log slow requests (>5s) for performance diagnostics
            if elapsed > 5.0:
                logger.warning(
                    "Slow request (%.2fs) to %s - may indicate proxy timeout risk",
                    elapsed,
                    endpoint,
                )
            elif elapsed > 2.0:
                logger.debug("Request to %s took %.2fs", endpoint, elapsed)
            return result
        except RetryableHTTPError as e:
            # Re-raise as HTTPStatusError to maintain backward compatibility
            elapsed = time.monotonic() - start_time
            logger.warning(
                "Request to %s failed after %.2fs",
                endpoint,
                elapsed,
            )
            e.response.raise_for_status()
            raise  # pragma: no cover  # Unreachable: raise_for_status always raises
        except Exception:
            elapsed = time.monotonic() - start_time
            logger.warning(
                "Request to %s failed after %.2fs",
                endpoint,
                elapsed,
            )
            raise

    def _log_retry(self, retry_state: Any) -> None:
        """Log retry attempts.

        Args:
            retry_state: Tenacity retry state object
        """
        logger.warning(
            "Retry attempt %d after error: %s",
            retry_state.attempt_number,
            retry_state.outcome.exception() if retry_state.outcome else "unknown",
        )

    async def invalidate_cache(self, endpoint: str, params: dict[str, Any] | None = None) -> bool:
        """Invalidate a specific cache entry.

        Args:
            endpoint: The API endpoint path
            params: Optional query parameters

        Returns:
            True if entry was found and removed, False otherwise
        """
        cache_key = self._make_cache_key(endpoint, params)
        async with self._cache_lock:
            if cache_key in self._cache:
                del self._cache[cache_key]
                return True
            return False

    async def clear_cache(self) -> int:
        """Clear all cached entries.

        Returns:
            The number of entries that were cleared
        """
        async with self._cache_lock:
            count = len(self._cache)
            self._cache.clear()
            return count

    async def _post(self, endpoint: str, json: dict[str, Any] | None = None) -> Any:
        """Make a POST request.

        Args:
            endpoint: The API endpoint path
            json: Optional JSON body

        Returns:
            The JSON response data

        Raises:
            httpx.HTTPStatusError: On HTTP errors (after retries exhausted)
        """
        return await self._request_with_retry("POST", endpoint, json=json)

    async def _put(self, endpoint: str, json: dict[str, Any] | None = None) -> Any:
        """Make a PUT request.

        Args:
            endpoint: The API endpoint path
            json: Optional JSON body

        Returns:
            The JSON response data

        Raises:
            httpx.HTTPStatusError: On HTTP errors (after retries exhausted)
        """
        return await self._request_with_retry("PUT", endpoint, json=json)

    @staticmethod
    def _parse_release(item: dict[str, Any]) -> Release:
        """Parse a release from API response.

        Args:
            item: A single release item from the API response

        Returns:
            A Release model instance
        """
        from filtarr.models.common import Quality, Release

        quality_data = item.get("quality", {}).get("quality", {})
        return Release(
            guid=item["guid"],
            title=item["title"],
            indexer=item.get("indexer", "Unknown"),
            size=item.get("size", 0),
            quality=Quality(
                id=quality_data.get("id", 0),
                name=quality_data.get("name", "Unknown"),
            ),
        )



"""Extended tests for CLI batch processing edge cases."""

from __future__ import annotations

from io import StringIO
from typing import TYPE_CHECKING, Any
from unittest.mock import AsyncMock, MagicMock, patch

import httpx
import pytest
import respx
from httpx import Response
from rich.console import Console
from typer.testing import CliRunner

from filtarr.checker import ReleaseChecker, SamplingStrategy, SearchResult
from filtarr.cli import (
    BatchContext,
    OutputFormat,
    _build_item_list,
    _fetch_movies_to_check,
    _fetch_series_to_check,
    _handle_batch_result,
    _process_batch_item,
    _process_movie_item,
    _process_series_item,
    _process_single_item,
    _run_batch_checks,
    app,
)
from filtarr.config import Config, ConfigurationError, RadarrConfig, SonarrConfig, TagConfig
from filtarr.criteria import SearchCriteria
from filtarr.state import BatchProgress, StateManager
from filtarr.tagger import TagResult

if TYPE_CHECKING:
    from filtarr.models.radarr import Movie
    from filtarr.models.sonarr import Series

runner = CliRunner()


@pytest.fixture
def mock_config() -> Config:
    """Create a mock config for testing."""
    return Config(
        radarr=RadarrConfig(url="http://localhost:7878", api_key="radarr-key"),
        sonarr=SonarrConfig(url="http://127.0.0.1:8989", api_key="sonarr-key"),
        timeout=30.0,
        tags=TagConfig(),
    )


@pytest.fixture
def mock_console() -> Any:
    """Create a mock console for testing."""
    console = MagicMock(spec=Console)
    console.print = MagicMock()
    return console


@pytest.fixture
def mock_error_console() -> Any:
    """Create a mock error console for testing."""
    console = MagicMock(spec=Console)
    console.print = MagicMock()
    return console


@pytest.fixture
def mock_state_manager() -> MagicMock:
    """Create a mock StateManager for testing."""
    mock = MagicMock(spec=StateManager)
    mock.record_check = MagicMock()
    mock.update_batch_progress = MagicMock()
    mock.start_batch = MagicMock()
    mock.get_batch_progress = MagicMock(return_value=None)
    mock.clear_batch_progress = MagicMock()
    mock.get_stale_unavailable_items = MagicMock(return_value=[])
    mock.get_cached_result = MagicMock(return_value=None)
    return mock


@pytest.fixture
def real_console() -> Console:
    """Create a real Console for tests that require it (Progress bar)."""
    return Console(file=StringIO(), force_terminal=True)


class TestFetchMoviesToCheck:
    """Tests for _fetch_movies_to_check function."""

    @respx.mock
    @pytest.mark.asyncio
    async def test_fetch_movies_skip_tagged_true_filters_movies(
        self, mock_config: Config, mock_console: Any
    ) -> None:
        """skip_tagged=True should filter out movies with matching tags."""
        respx.get("http://localhost:7878/api/v3/tag").mock(
            return_value=Response(
                200,
                json=[
                    {"id": 1, "label": "4k-available"},
                    {"id": 2, "label": "4k-unavailable"},
                ],
            )
        )
        respx.get("http://localhost:7878/api/v3/movie").mock(
            return_value=Response(
                200,
                json=[
                    {"id": 1, "title": "Tagged Movie", "year": 2024, "tags": [1]},
                    {"id": 2, "title": "Untagged Movie", "year": 2024, "tags": []},
                    {"id": 3, "title": "Other Tagged Movie", "year": 2024, "tags": [2]},
                ],
            )
        )

        movies, skip_tags = await _fetch_movies_to_check(
            mock_config, SearchCriteria.FOUR_K, skip_tagged=True, console=mock_console
        )

        # Only movie 2 should be in the result (movies 1 and 3 have skip tags)
        assert len(movies) == 1
        assert movies[0].id == 2
        assert skip_tags == {1, 2}

    @respx.mock
    @pytest.mark.asyncio
    async def test_fetch_movies_skip_tagged_false_returns_all(
        self, mock_config: Config, mock_console: Any
    ) -> None:
        """skip_tagged=False should return all movies."""
        respx.get("http://localhost:7878/api/v3/movie").mock(
            return_value=Response(
                200,
                json=[
                    {"id": 1, "title": "Tagged Movie", "year": 2024, "tags": [1]},
                    {"id": 2, "title": "Untagged Movie", "year": 2024, "tags": []},
                ],
            )
        )

        movies, skip_tags = await _fetch_movies_to_check(
            mock_config, SearchCriteria.FOUR_K, skip_tagged=False, console=mock_console
        )

        # All movies should be returned
        assert len(movies) == 2
        assert skip_tags == set()

    @respx.mock
    @pytest.mark.asyncio
    async def test_fetch_movies_with_different_criteria_tag_names(
        self, mock_config: Config, mock_console: Any
    ) -> None:
        """Different criteria should use different tag names for filtering."""
        respx.get("http://localhost:7878/api/v3/tag").mock(
            return_value=Response(
                200,
                json=[
                    {"id": 1, "label": "4k-available"},
                    {"id": 2, "label": "hdr-available"},
                    {"id": 3, "label": "hdr-unavailable"},
                ],
            )
        )
        respx.get("http://localhost:7878/api/v3/movie").mock(
            return_value=Response(
                200,
                json=[
                    {"id": 1, "title": "4K Movie", "year": 2024, "tags": [1]},  # Has 4k tag
                    {"id": 2, "title": "HDR Movie", "year": 2024, "tags": [2]},  # Has hdr tag
                    {"id": 3, "title": "Fresh Movie", "year": 2024, "tags": []},
                ],
            )
        )

        # Search for HDR - should skip movie 2 (has hdr tag) but include movie 1
        movies, skip_tags = await _fetch_movies_to_check(
            mock_config, SearchCriteria.HDR, skip_tagged=True, console=mock_console
        )

        assert len(movies) == 2  # Movies 1 and 3 (movie 2 has hdr tag)
        assert skip_tags == {2, 3}


class TestFetchSeriesToCheck:
    """Tests for _fetch_series_to_check function."""

    @respx.mock
    @pytest.mark.asyncio
    async def test_fetch_series_skip_tagged_true_filters_series(
        self, mock_config: Config, mock_console: Any
    ) -> None:
        """skip_tagged=True should filter out series with matching tags."""
        respx.get("http://127.0.0.1:8989/api/v3/tag").mock(
            return_value=Response(
                200,
                json=[
                    {"id": 1, "label": "4k-available"},
                    {"id": 2, "label": "4k-unavailable"},
                ],
            )
        )
        respx.get("http://127.0.0.1:8989/api/v3/series").mock(
            return_value=Response(
                200,
                json=[
                    {"id": 1, "title": "Tagged Series", "year": 2024, "seasons": [], "tags": [1]},
                    {"id": 2, "title": "Untagged Series", "year": 2024, "seasons": [], "tags": []},
                ],
            )
        )

        series, skip_tags = await _fetch_series_to_check(
            mock_config, SearchCriteria.FOUR_K, skip_tagged=True, console=mock_console
        )

        # Only series 2 should be in result
        assert len(series) == 1
        assert series[0].id == 2
        assert skip_tags == {1, 2}

    @respx.mock
    @pytest.mark.asyncio
    async def test_fetch_series_skip_tagged_false_returns_all(
        self, mock_config: Config, mock_console: Any
    ) -> None:
        """skip_tagged=False should return all series."""
        respx.get("http://127.0.0.1:8989/api/v3/series").mock(
            return_value=Response(
                200,
                json=[
                    {"id": 1, "title": "Tagged Series", "year": 2024, "seasons": [], "tags": [1]},
                    {"id": 2, "title": "Untagged Series", "year": 2024, "seasons": [], "tags": []},
                ],
            )
        )

        series, skip_tags = await _fetch_series_to_check(
            mock_config, SearchCriteria.FOUR_K, skip_tagged=False, console=mock_console
        )

        # All series should be returned
        assert len(series) == 2
        assert skip_tags == set()


class TestProcessMovieItem:
    """Tests for _process_movie_item function."""

    @respx.mock
    @pytest.mark.asyncio
    async def test_process_movie_by_name_no_matches(
        self, mock_console: Any, mock_error_console: Any
    ) -> None:
        """Movie lookup by name with no matches should return None."""
        respx.get("http://localhost:7878/api/v3/movie").mock(return_value=Response(200, json=[]))

        checker = ReleaseChecker(radarr_url="http://localhost:7878", radarr_api_key="test")

        result = await _process_movie_item(
            checker=checker,
            item_id=-1,  # Negative means lookup by name
            item_name="Nonexistent Movie",
            search_criteria=SearchCriteria.FOUR_K,
            apply_tags=False,
            dry_run=False,
            console=mock_console,
            error_console=mock_error_console,
        )

        assert result is None
        mock_error_console.print.assert_called()
        # Check that error was printed
        call_args = str(mock_error_console.print.call_args)
        assert "Movie not found" in call_args or "not found" in call_args.lower()

    @respx.mock
    @pytest.mark.asyncio
    async def test_process_movie_by_name_multiple_matches(
        self, mock_console: Any, mock_error_console: Any
    ) -> None:
        """Movie lookup by name with multiple matches should return None."""
        respx.get("http://localhost:7878/api/v3/movie").mock(
            return_value=Response(
                200,
                json=[
                    {"id": 1, "title": "The Matrix", "year": 1999},
                    {"id": 2, "title": "The Matrix Reloaded", "year": 2003},
                    {"id": 3, "title": "The Matrix Revolutions", "year": 2003},
                ],
            )
        )

        checker = ReleaseChecker(radarr_url="http://localhost:7878", radarr_api_key="test")

        result = await _process_movie_item(
            checker=checker,
            item_id=-1,
            item_name="The Matrix",
            search_criteria=SearchCriteria.FOUR_K,
            apply_tags=False,
            dry_run=False,
            console=mock_console,
            error_console=mock_error_console,
        )

        assert result is None
        mock_error_console.print.assert_called()
        call_args = str(mock_error_console.print.call_args)
        assert "Multiple movies" in call_args or "multiple" in call_args.lower()

    @respx.mock
    @pytest.mark.asyncio
    async def test_process_movie_by_id_success(
        self, mock_console: Any, mock_error_console: Any
    ) -> None:
        """Movie lookup by ID should check the movie directly."""
        respx.get("http://localhost:7878/api/v3/movie/123").mock(
            return_value=Response(
                200,
                json={"id": 123, "title": "Test Movie", "year": 2024, "tags": []},
            )
        )
        respx.get("http://localhost:7878/api/v3/release", params={"movieId": "123"}).mock(
            return_value=Response(
                200,
                json=[
                    {
                        "guid": "rel1",
                        "title": "Movie.2160p.BluRay",
                        "indexer": "Test",
                        "size": 5000,
                        "quality": {"quality": {"id": 19, "name": "WEBDL-2160p"}},
                    }
                ],
            )
        )

        checker = ReleaseChecker(radarr_url="http://localhost:7878", radarr_api_key="test")

        result = await _process_movie_item(
            checker=checker,
            item_id=123,
            item_name="Test Movie",
            search_criteria=SearchCriteria.FOUR_K,
            apply_tags=False,
            dry_run=False,
            console=mock_console,
            error_console=mock_error_console,
        )

        assert result is not None
        assert result.has_match is True
        assert result.item_id == 123

    @respx.mock
    @pytest.mark.asyncio
    async def test_process_movie_by_name_single_match(
        self, mock_console: Any, mock_error_console: Any
    ) -> None:
        """Movie lookup by name with single match should process the movie."""
        respx.get("http://localhost:7878/api/v3/movie").mock(
            return_value=Response(
                200,
                json=[{"id": 456, "title": "Inception", "year": 2010}],
            )
        )
        # check_movie fetches movie info first
        respx.get("http://localhost:7878/api/v3/movie/456").mock(
            return_value=Response(
                200,
                json={"id": 456, "title": "Inception", "year": 2010, "tags": []},
            )
        )
        respx.get("http://localhost:7878/api/v3/release", params={"movieId": "456"}).mock(
            return_value=Response(200, json=[])
        )

        checker = ReleaseChecker(radarr_url="http://localhost:7878", radarr_api_key="test")

        result = await _process_movie_item(
            checker=checker,
            item_id=-1,
            item_name="Inception",
            search_criteria=SearchCriteria.FOUR_K,
            apply_tags=False,
            dry_run=False,
            console=mock_console,
            error_console=mock_error_console,
        )

        assert result is not None
        assert result.item_id == 456
        mock_console.print.assert_called()  # "Found: Inception" message


class TestProcessSeriesItem:
    """Tests for _process_series_item function."""

    @respx.mock
    @pytest.mark.asyncio
    async def test_process_series_by_name_no_matches(
        self, mock_console: Any, mock_error_console: Any
    ) -> None:
        """Series lookup by name with no matches should return None."""
        respx.get("http://127.0.0.1:8989/api/v3/series").mock(return_value=Response(200, json=[]))

        checker = ReleaseChecker(sonarr_url="http://127.0.0.1:8989", sonarr_api_key="test")

        result = await _process_series_item(
            checker=checker,
            item_id=-1,
            item_name="Nonexistent Series",
            series_criteria=SearchCriteria.FOUR_K,
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            console=mock_console,
            error_console=mock_error_console,
        )

        assert result is None
        mock_error_console.print.assert_called()
        call_args = str(mock_error_console.print.call_args)
        assert "Series not found" in call_args or "not found" in call_args.lower()

    @respx.mock
    @pytest.mark.asyncio
    async def test_process_series_by_name_multiple_matches(
        self, mock_console: Any, mock_error_console: Any
    ) -> None:
        """Series lookup by name with multiple matches should return None."""
        respx.get("http://127.0.0.1:8989/api/v3/series").mock(
            return_value=Response(
                200,
                json=[
                    {"id": 1, "title": "Breaking Bad", "year": 2008, "seasons": []},
                    {"id": 2, "title": "Breaking Bad (2022)", "year": 2022, "seasons": []},
                ],
            )
        )

        checker = ReleaseChecker(sonarr_url="http://127.0.0.1:8989", sonarr_api_key="test")

        result = await _process_series_item(
            checker=checker,
            item_id=-1,
            item_name="Breaking",
            series_criteria=SearchCriteria.FOUR_K,
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            console=mock_console,
            error_console=mock_error_console,
        )

        assert result is None
        mock_error_console.print.assert_called()
        call_args = str(mock_error_console.print.call_args)
        assert "Multiple series" in call_args or "multiple" in call_args.lower()

    @respx.mock
    @pytest.mark.asyncio
    async def test_process_series_by_id_success(
        self, mock_console: Any, mock_error_console: Any
    ) -> None:
        """Series lookup by ID should check the series directly."""
        respx.get("http://127.0.0.1:8989/api/v3/series/789").mock(
            return_value=Response(
                200,
                json={"id": 789, "title": "Test Series", "year": 2020, "seasons": [], "tags": []},
            )
        )
        respx.get("http://127.0.0.1:8989/api/v3/episode", params={"seriesId": "789"}).mock(
            return_value=Response(
                200,
                json=[
                    {
                        "id": 101,
                        "seriesId": 789,
                        "seasonNumber": 1,
                        "episodeNumber": 1,
                        "airDate": "2020-01-01",
                        "monitored": True,
                    }
                ],
            )
        )
        respx.get("http://127.0.0.1:8989/api/v3/release", params={"episodeId": "101"}).mock(
            return_value=Response(
                200,
                json=[
                    {
                        "guid": "rel-101",
                        "title": "Series.S01E01.2160p.WEB-DL",
                        "indexer": "Test",
                        "size": 3000,
                        "quality": {"quality": {"id": 19, "name": "WEBDL-2160p"}},
                    }
                ],
            )
        )

        checker = ReleaseChecker(sonarr_url="http://127.0.0.1:8989", sonarr_api_key="test")

        result = await _process_series_item(
            checker=checker,
            item_id=789,
            item_name="Test Series",
            series_criteria=SearchCriteria.FOUR_K,
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            console=mock_console,
            error_console=mock_error_console,
        )

        assert result is not None
        assert result.has_match is True
        assert result.item_id == 789

    @respx.mock
    @pytest.mark.asyncio
    async def test_process_series_by_name_single_match(
        self, mock_console: Any, mock_error_console: Any
    ) -> None:
        """Series lookup by name with single match should process the series."""
        respx.get("http://127.0.0.1:8989/api/v3/series").mock(
            return_value=Response(
                200,
                json=[{"id": 456, "title": "Game of Thrones", "year": 2011, "seasons": []}],
            )
        )
        respx.get("http://127.0.0.1:8989/api/v3/series/456").mock(
            return_value=Response(
                200,
                json={"id": 456, "title": "Game of Thrones", "year": 2011, "seasons": []},
            )
        )
        respx.get("http://127.0.0.1:8989/api/v3/episode", params={"seriesId": "456"}).mock(
            return_value=Response(
                200,
                json=[
                    {
                        "id": 1001,
                        "seriesId": 456,
                        "seasonNumber": 1,
                        "episodeNumber": 1,
                        "airDate": "2011-04-17",
                        "monitored": True,
                    }
                ],
            )
        )
        respx.get("http://127.0.0.1:8989/api/v3/release", params={"episodeId": "1001"}).mock(
            return_value=Response(200, json=[])
        )

        checker = ReleaseChecker(sonarr_url="http://127.0.0.1:8989", sonarr_api_key="test")

        result = await _process_series_item(
            checker=checker,
            item_id=-1,
            item_name="Game of Thrones",
            series_criteria=SearchCriteria.FOUR_K,
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            console=mock_console,
            error_console=mock_error_console,
        )

        assert result is not None
        assert result.item_id == 456
        mock_console.print.assert_called()  # "Found: Game of Thrones" message


class TestProcessMovieItemWithDifferentCriteria:
    """Tests for _process_movie_item with various criteria."""

    @respx.mock
    @pytest.mark.asyncio
    async def test_process_movie_with_hdr_criteria(
        self, mock_console: Any, mock_error_console: Any
    ) -> None:
        """Movie processing with HDR criteria should search for HDR releases."""
        respx.get("http://localhost:7878/api/v3/movie/123").mock(
            return_value=Response(
                200,
                json={"id": 123, "title": "HDR Movie", "year": 2024, "tags": []},
            )
        )
        respx.get("http://localhost:7878/api/v3/release", params={"movieId": "123"}).mock(
            return_value=Response(
                200,
                json=[
                    {
                        "guid": "rel1",
                        "title": "Movie.2160p.HDR.BluRay",
                        "indexer": "Test",
                        "size": 5000,
                        "quality": {"quality": {"id": 31, "name": "Bluray-2160p"}},
                    }
                ],
            )
        )

        checker = ReleaseChecker(radarr_url="http://localhost:7878", radarr_api_key="test")

        result = await _process_movie_item(
            checker=checker,
            item_id=123,
            item_name="HDR Movie",
            search_criteria=SearchCriteria.HDR,
            apply_tags=False,
            dry_run=False,
            console=mock_console,
            error_console=mock_error_console,
        )

        assert result is not None
        assert result.has_match is True


class TestProcessSeriesItemWithStrategies:
    """Tests for _process_series_item with different sampling strategies."""

    @respx.mock
    @pytest.mark.asyncio
    async def test_process_series_with_all_strategy(
        self, mock_console: Any, mock_error_console: Any
    ) -> None:
        """Series processing with ALL strategy should check all seasons."""
        respx.get("http://127.0.0.1:8989/api/v3/series/123").mock(
            return_value=Response(
                200,
                json={"id": 123, "title": "Short Series", "year": 2023, "seasons": []},
            )
        )
        respx.get("http://127.0.0.1:8989/api/v3/episode", params={"seriesId": "123"}).mock(
            return_value=Response(
                200,
                json=[
                    {
                        "id": 101,
                        "seriesId": 123,
                        "seasonNumber": 1,
                        "episodeNumber": 1,
                        "airDate": "2023-01-01",
                        "monitored": True,
                    },
                    {
                        "id": 201,
                        "seriesId": 123,
                        "seasonNumber": 2,
                        "episodeNumber": 1,
                        "airDate": "2023-06-01",
                        "monitored": True,
                    },
                ],
            )
        )
        respx.get("http://127.0.0.1:8989/api/v3/release", params={"episodeId": "101"}).mock(
            return_value=Response(200, json=[])
        )
        respx.get("http://127.0.0.1:8989/api/v3/release", params={"episodeId": "201"}).mock(
            return_value=Response(200, json=[])
        )

        checker = ReleaseChecker(sonarr_url="http://127.0.0.1:8989", sonarr_api_key="test")

        result = await _process_series_item(
            checker=checker,
            item_id=123,
            item_name="Short Series",
            series_criteria=SearchCriteria.FOUR_K,
            sampling_strategy=SamplingStrategy.ALL,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            console=mock_console,
            error_console=mock_error_console,
        )

        assert result is not None
        # ALL strategy should check all seasons (2 in this case)
        assert len(result.seasons_checked) == 2

    @respx.mock
    @pytest.mark.asyncio
    async def test_process_series_with_distributed_strategy(
        self, mock_console: Any, mock_error_console: Any
    ) -> None:
        """Series processing with DISTRIBUTED strategy should check first, middle, last."""
        respx.get("http://127.0.0.1:8989/api/v3/series/456").mock(
            return_value=Response(
                200,
                json={"id": 456, "title": "Long Series", "year": 2018, "seasons": []},
            )
        )
        respx.get("http://127.0.0.1:8989/api/v3/episode", params={"seriesId": "456"}).mock(
            return_value=Response(
                200,
                json=[
                    {
                        "id": 100 + i,
                        "seriesId": 456,
                        "seasonNumber": i,
                        "episodeNumber": 1,
                        "airDate": f"201{i}-01-01",
                        "monitored": True,
                    }
                    for i in range(1, 6)  # Seasons 1-5
                ],
            )
        )
        # Distributed should check seasons 1, 3, 5
        for ep_id in [101, 103, 105]:
            respx.get(
                "http://127.0.0.1:8989/api/v3/release", params={"episodeId": str(ep_id)}
            ).mock(return_value=Response(200, json=[]))

        checker = ReleaseChecker(sonarr_url="http://127.0.0.1:8989", sonarr_api_key="test")

        result = await _process_series_item(
            checker=checker,
            item_id=456,
            item_name="Long Series",
            series_criteria=SearchCriteria.FOUR_K,
            sampling_strategy=SamplingStrategy.DISTRIBUTED,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            console=mock_console,
            error_console=mock_error_console,
        )

        assert result is not None
        # DISTRIBUTED should check first, middle, last (seasons 1, 3, 5)
        assert sorted(result.seasons_checked) == [1, 3, 5]


class TestProcessItemsWithDryRun:
    """Tests for dry_run mode in item processing."""

    @respx.mock
    @pytest.mark.asyncio
    async def test_process_movie_dry_run_no_tag_changes(
        self, mock_console: Any, mock_error_console: Any
    ) -> None:
        """dry_run mode should not apply actual tags."""
        respx.get("http://localhost:7878/api/v3/movie/123").mock(
            return_value=Response(
                200,
                json={"id": 123, "title": "Test Movie", "year": 2024, "tags": []},
            )
        )
        respx.get("http://localhost:7878/api/v3/release", params={"movieId": "123"}).mock(
            return_value=Response(
                200,
                json=[
                    {
                        "guid": "rel1",
                        "title": "Movie.2160p.BluRay",
                        "indexer": "Test",
                        "size": 5000,
                        "quality": {"quality": {"id": 19, "name": "WEBDL-2160p"}},
                    }
                ],
            )
        )
        # No tag API mocks needed - dry_run shouldn't call them

        checker = ReleaseChecker(
            radarr_url="http://localhost:7878",
            radarr_api_key="test",
            tag_config=TagConfig(),
        )

        result = await _process_movie_item(
            checker=checker,
            item_id=123,
            item_name="Test Movie",
            search_criteria=SearchCriteria.FOUR_K,
            apply_tags=True,
            dry_run=True,
            console=mock_console,
            error_console=mock_error_console,
        )

        assert result is not None
        assert result.has_match is True
        assert result.tag_result is not None
        assert result.tag_result.dry_run is True


class TestProcessMovieItemEdgeCases:
    """Edge case tests for _process_movie_item."""

    @respx.mock
    @pytest.mark.asyncio
    async def test_process_movie_by_name_truncates_multiple_matches_display(
        self, mock_console: Any, mock_error_console: Any
    ) -> None:
        """When many matches are found, display should be truncated."""
        respx.get("http://localhost:7878/api/v3/movie").mock(
            return_value=Response(
                200,
                json=[
                    {"id": i, "title": f"Movie Part {i}", "year": 2020 + i}
                    for i in range(1, 10)  # 9 movies
                ],
            )
        )

        checker = ReleaseChecker(radarr_url="http://localhost:7878", radarr_api_key="test")

        result = await _process_movie_item(
            checker=checker,
            item_id=-1,
            item_name="Movie Part",
            search_criteria=SearchCriteria.FOUR_K,
            apply_tags=False,
            dry_run=False,
            console=mock_console,
            error_console=mock_error_console,
        )

        assert result is None
        # Error console should show truncated message with "..."
        call_args = str(mock_error_console.print.call_args)
        assert "..." in call_args


class TestProcessSeriesItemEdgeCases:
    """Edge case tests for _process_series_item."""

    @respx.mock
    @pytest.mark.asyncio
    async def test_process_series_by_name_truncates_multiple_matches_display(
        self, mock_console: Any, mock_error_console: Any
    ) -> None:
        """When many series matches are found, display should be truncated."""
        respx.get("http://127.0.0.1:8989/api/v3/series").mock(
            return_value=Response(
                200,
                json=[
                    {"id": i, "title": f"Series Season {i}", "year": 2020 + i, "seasons": []}
                    for i in range(1, 10)  # 9 series
                ],
            )
        )

        checker = ReleaseChecker(sonarr_url="http://127.0.0.1:8989", sonarr_api_key="test")

        result = await _process_series_item(
            checker=checker,
            item_id=-1,
            item_name="Series Season",
            series_criteria=SearchCriteria.FOUR_K,
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            console=mock_console,
            error_console=mock_error_console,
        )

        assert result is None
        # Error console should show truncated message with "..."
        call_args = str(mock_error_console.print.call_args)
        assert "..." in call_args


# =============================================================================
# NEW TESTS FOR UNCOVERED LINES
# =============================================================================


class TestMovieOnlyCriteriaWarningForSeries:
    """Tests for L825-829: Movie-only criteria warning for series in batch processing."""

    @pytest.mark.asyncio
    async def test_process_batch_item_with_directors_cut_on_series(
        self, mock_config: Config, mock_console: Any, mock_error_console: Any
    ) -> None:
        """Should warn and fallback to 4K when directors-cut is used for series."""
        ctx = BatchContext(
            config=mock_config,
            state_manager=MagicMock(spec=StateManager),
            search_criteria=SearchCriteria.DIRECTORS_CUT,
            criteria_str="directors-cut",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=mock_console,
            error_console=mock_error_console,
        )

        # Mock the get_checker to return an async mock checker
        mock_checker = AsyncMock()
        mock_checker.check_series.return_value = SearchResult(
            item_id=456, item_type="series", has_match=False
        )

        with patch("filtarr.cli.get_checker", return_value=mock_checker):
            await _process_batch_item(ctx, "series", 456, "Test Series")

        # Should have printed the warning about movie-only criteria
        warning_printed = any(
            "directors-cut" in str(call).lower() and "movie-only" in str(call).lower()
            for call in mock_console.print.call_args_list
        )
        assert warning_printed or any(
            "Using 4K" in str(call) for call in mock_console.print.call_args_list
        )

    @pytest.mark.asyncio
    async def test_process_batch_item_with_extended_on_series(
        self, mock_config: Config, mock_console: Any, mock_error_console: Any
    ) -> None:
        """Should warn and fallback to 4K when extended is used for series."""
        ctx = BatchContext(
            config=mock_config,
            state_manager=MagicMock(spec=StateManager),
            search_criteria=SearchCriteria.EXTENDED,
            criteria_str="extended",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=mock_console,
            error_console=mock_error_console,
        )

        mock_checker = AsyncMock()
        mock_checker.check_series.return_value = SearchResult(
            item_id=456, item_type="series", has_match=False
        )

        with patch("filtarr.cli.get_checker", return_value=mock_checker):
            await _process_batch_item(ctx, "series", 456, "Test Series")

        # Check that warning about movie-only criteria was printed
        assert mock_console.print.called
        call_args_str = " ".join(str(call) for call in mock_console.print.call_args_list)
        assert "4K" in call_args_str or "movie-only" in call_args_str.lower()


class TestHandleBatchResult:
    """Tests for L857, L866: State recording and batch progress update."""

    def test_handle_batch_result_records_state_when_not_dry_run_with_tags(
        self,
        mock_config: Config,
        mock_console: Any,
        mock_error_console: Any,
        mock_state_manager: MagicMock,
    ) -> None:
        """Should record check in state when dry_run=False and apply_tags=True."""
        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=True,
            dry_run=False,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=mock_console,
            error_console=mock_error_console,
        )

        # Create a result with tag_result
        result = SearchResult(
            item_id=123,
            item_type="movie",
            has_match=True,
            tag_result=TagResult(tag_applied="4k-available"),
        )

        batch_progress = BatchProgress(batch_id="test-batch", item_type="movie", total_items=10)

        _handle_batch_result(ctx, result, item_id=123, batch_progress=batch_progress)

        # Should have recorded the check
        mock_state_manager.record_check.assert_called_once_with("movie", 123, True, "4k-available")
        # Should have updated batch progress
        mock_state_manager.update_batch_progress.assert_called_once_with(123)

    def test_handle_batch_result_updates_progress_with_valid_item_id(
        self,
        mock_config: Config,
        mock_console: Any,
        mock_error_console: Any,
        mock_state_manager: MagicMock,
    ) -> None:
        """Should update batch progress when item_id > 0."""
        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=True,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=mock_console,
            error_console=mock_error_console,
        )

        result = SearchResult(item_id=456, item_type="movie", has_match=False)
        batch_progress = BatchProgress(batch_id="test-batch", item_type="movie", total_items=10)

        _handle_batch_result(ctx, result, item_id=456, batch_progress=batch_progress)

        # Should have updated batch progress (even in dry_run)
        mock_state_manager.update_batch_progress.assert_called_once_with(456)

    def test_handle_batch_result_no_update_for_negative_item_id(
        self,
        mock_config: Config,
        mock_console: Any,
        mock_error_console: Any,
        mock_state_manager: MagicMock,
    ) -> None:
        """Should not update batch progress when item_id <= 0."""
        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=mock_console,
            error_console=mock_error_console,
        )

        result = SearchResult(item_id=123, item_type="movie", has_match=False)
        batch_progress = BatchProgress(batch_id="test-batch", item_type="movie", total_items=10)

        _handle_batch_result(ctx, result, item_id=-1, batch_progress=batch_progress)

        # Should NOT have updated batch progress for negative item_id
        mock_state_manager.update_batch_progress.assert_not_called()


class TestBuildItemList:
    """Tests for L880, L883: Movie/series append in _build_item_list()."""

    def test_build_item_list_with_movies(self) -> None:
        """Should add movies to all_items list."""
        # Create mock Movie objects
        mock_movie1 = MagicMock()
        mock_movie1.id = 1
        mock_movie1.title = "Movie 1"

        mock_movie2 = MagicMock()
        mock_movie2.id = 2
        mock_movie2.title = "Movie 2"

        movies: list[Movie] = [mock_movie1, mock_movie2]
        series: list[Series] = []
        file_items: list[tuple[str, str]] = []

        result = _build_item_list(movies, series, file_items)

        assert len(result) == 2
        assert result[0] == ("movie", 1, "Movie 1")
        assert result[1] == ("movie", 2, "Movie 2")

    def test_build_item_list_with_series(self) -> None:
        """Should add series to all_items list."""
        mock_series1 = MagicMock()
        mock_series1.id = 100
        mock_series1.title = "Series 1"

        mock_series2 = MagicMock()
        mock_series2.id = 200
        mock_series2.title = "Series 2"

        movies: list[Movie] = []
        series: list[Series] = [mock_series1, mock_series2]
        file_items: list[tuple[str, str]] = []

        result = _build_item_list(movies, series, file_items)

        assert len(result) == 2
        assert result[0] == ("series", 100, "Series 1")
        assert result[1] == ("series", 200, "Series 2")

    def test_build_item_list_with_mixed_items(self) -> None:
        """Should add movies, series, and file items correctly."""
        mock_movie = MagicMock()
        mock_movie.id = 1
        mock_movie.title = "Test Movie"

        mock_series = MagicMock()
        mock_series.id = 100
        mock_series.title = "Test Series"

        movies: list[Movie] = [mock_movie]
        series: list[Series] = [mock_series]
        file_items: list[tuple[str, str]] = [
            ("movie", "456"),
            ("series", "Breaking Bad"),
        ]

        result = _build_item_list(movies, series, file_items)

        assert len(result) == 4
        assert result[0] == ("movie", 1, "Test Movie")
        assert result[1] == ("series", 100, "Test Series")
        assert result[2] == ("movie", 456, "ID:456")
        assert result[3] == ("series", -1, "Breaking Bad")


class TestProcessSingleItemErrorHandlers:
    """Tests for L945-962: Batch error handlers in _process_single_item()."""

    @pytest.mark.asyncio
    async def test_batch_limit_reached_returns_false(
        self,
        mock_config: Config,
        mock_console: Any,
        mock_error_console: Any,
        mock_state_manager: MagicMock,
    ) -> None:
        """Should return False and set batch_limit_reached when batch_size > 0 and limit reached."""
        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=1,  # Limit of 1
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=mock_console,
            error_console=mock_error_console,
        )

        mock_result = SearchResult(item_id=123, item_type="movie", has_match=True)

        with patch("filtarr.cli._process_batch_item", return_value=mock_result):
            should_continue = await _process_single_item(
                ctx, "movie", 123, "Test Movie", batch_progress=None
            )

        # First item should be processed (processed_this_run=1), then limit reached
        assert ctx.processed_this_run == 1
        assert ctx.batch_limit_reached is True
        assert should_continue is False

    @pytest.mark.asyncio
    async def test_configuration_error_handling(
        self,
        mock_config: Config,
        mock_console: Any,
        mock_error_console: Any,
        mock_state_manager: MagicMock,
    ) -> None:
        """Should handle ConfigurationError and mark as processed."""
        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=mock_console,
            error_console=mock_error_console,
        )

        batch_progress = BatchProgress(batch_id="test", item_type="movie", total_items=10)

        with patch(
            "filtarr.cli._process_batch_item",
            side_effect=ConfigurationError("Radarr not configured"),
        ):
            should_continue = await _process_single_item(
                ctx, "movie", 123, "Test Movie", batch_progress=batch_progress
            )

        # Should print error and mark as processed
        assert should_continue is True
        mock_error_console.print.assert_called()
        mock_state_manager.update_batch_progress.assert_called_once_with(123)

    @pytest.mark.asyncio
    async def test_http_status_error_transient_not_marked_processed(
        self,
        mock_config: Config,
        mock_console: Any,
        mock_error_console: Any,
        mock_state_manager: MagicMock,
    ) -> None:
        """Should NOT mark as processed for transient HTTP errors (5xx)."""
        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=mock_console,
            error_console=mock_error_console,
        )

        batch_progress = BatchProgress(batch_id="test", item_type="movie", total_items=10)

        mock_response = MagicMock()
        mock_response.status_code = 502  # Bad Gateway - transient
        mock_request = MagicMock()

        with patch(
            "filtarr.cli._process_batch_item",
            side_effect=httpx.HTTPStatusError(
                "Bad Gateway", request=mock_request, response=mock_response
            ),
        ):
            should_continue = await _process_single_item(
                ctx, "movie", 123, "Test Movie", batch_progress=batch_progress
            )

        # Should print error but NOT mark as processed for transient error
        assert should_continue is True
        mock_error_console.print.assert_called()
        mock_state_manager.update_batch_progress.assert_not_called()

    @pytest.mark.asyncio
    async def test_http_status_error_non_transient_marked_processed(
        self,
        mock_config: Config,
        mock_console: Any,
        mock_error_console: Any,
        mock_state_manager: MagicMock,
    ) -> None:
        """Should mark as processed for non-transient HTTP errors (4xx)."""
        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=mock_console,
            error_console=mock_error_console,
        )

        batch_progress = BatchProgress(batch_id="test", item_type="movie", total_items=10)

        mock_response = MagicMock()
        mock_response.status_code = 404  # Not Found - permanent
        mock_request = MagicMock()

        with patch(
            "filtarr.cli._process_batch_item",
            side_effect=httpx.HTTPStatusError(
                "Not Found", request=mock_request, response=mock_response
            ),
        ):
            should_continue = await _process_single_item(
                ctx, "movie", 123, "Test Movie", batch_progress=batch_progress
            )

        # Should print error and mark as processed
        assert should_continue is True
        mock_error_console.print.assert_called()
        mock_state_manager.update_batch_progress.assert_called_once_with(123)

    @pytest.mark.asyncio
    async def test_network_error_not_marked_processed(
        self,
        mock_config: Config,
        mock_console: Any,
        mock_error_console: Any,
        mock_state_manager: MagicMock,
    ) -> None:
        """Should NOT mark as processed for network errors (transient)."""
        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=mock_console,
            error_console=mock_error_console,
        )

        batch_progress = BatchProgress(batch_id="test", item_type="movie", total_items=10)

        with patch(
            "filtarr.cli._process_batch_item",
            side_effect=httpx.ConnectError("Connection refused"),
        ):
            should_continue = await _process_single_item(
                ctx, "movie", 123, "Test Movie", batch_progress=batch_progress
            )

        # Should print error but NOT mark as processed
        assert should_continue is True
        mock_error_console.print.assert_called()
        mock_state_manager.update_batch_progress.assert_not_called()

    @pytest.mark.asyncio
    async def test_timeout_exception_not_marked_processed(
        self,
        mock_config: Config,
        mock_console: Any,
        mock_error_console: Any,
        mock_state_manager: MagicMock,
    ) -> None:
        """Should NOT mark as processed for timeout exceptions."""
        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=mock_console,
            error_console=mock_error_console,
        )

        batch_progress = BatchProgress(batch_id="test", item_type="movie", total_items=10)

        with patch(
            "filtarr.cli._process_batch_item",
            side_effect=httpx.TimeoutException("Request timed out"),
        ):
            should_continue = await _process_single_item(
                ctx, "movie", 123, "Test Movie", batch_progress=batch_progress
            )

        # Should print error but NOT mark as processed
        assert should_continue is True
        mock_error_console.print.assert_called()
        mock_state_manager.update_batch_progress.assert_not_called()

    @pytest.mark.asyncio
    async def test_general_exception_not_marked_processed(
        self,
        mock_config: Config,
        mock_console: Any,
        mock_error_console: Any,
        mock_state_manager: MagicMock,
    ) -> None:
        """Should NOT mark as processed for general exceptions."""
        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=mock_console,
            error_console=mock_error_console,
        )

        batch_progress = BatchProgress(batch_id="test", item_type="movie", total_items=10)

        with patch(
            "filtarr.cli._process_batch_item",
            side_effect=RuntimeError("Unknown error"),
        ):
            should_continue = await _process_single_item(
                ctx, "movie", 123, "Test Movie", batch_progress=batch_progress
            )

        # Should print error but NOT mark as processed
        assert should_continue is True
        mock_error_console.print.assert_called()
        mock_state_manager.update_batch_progress.assert_not_called()


class TestBatchProgressHandling:
    """Tests for L1002-1006: Existing batch progress handling and L1024-1026: Skip processed items."""

    def test_batch_command_uses_existing_progress_on_resume(self) -> None:
        """Should use existing batch progress when resume=True."""
        existing_progress = BatchProgress(
            batch_id="existing-batch",
            item_type="movie",
            total_items=10,
            processed_ids={1, 2, 3},
        )

        mock_config = Config(radarr=RadarrConfig(url="http://localhost:7878", api_key="key"))
        mock_state_manager = MagicMock(spec=StateManager)
        mock_state_manager.get_batch_progress.return_value = existing_progress
        mock_state_manager.get_stale_unavailable_items.return_value = []

        with (
            patch("filtarr.cli.Config.load", return_value=mock_config),
            patch("filtarr.cli.get_state_manager", return_value=mock_state_manager),
            patch("filtarr.cli._fetch_movies_to_check") as mock_fetch,
        ):
            mock_movie = MagicMock()
            mock_movie.id = 1
            mock_movie.title = "Already Processed Movie"
            mock_fetch.return_value = ([mock_movie], set())

            result = runner.invoke(
                app,
                ["check", "batch", "--all-movies", "--format", "simple"],
            )

        # Should have printed resuming message
        assert "Resuming batch" in result.output or "already processed" in result.output.lower()

    def test_batch_skips_processed_items_in_resume_mode(self) -> None:
        """Should skip items that are already processed in resume mode."""
        existing_progress = BatchProgress(
            batch_id="existing-batch",
            item_type="movie",
            total_items=3,
            processed_ids={1, 2},  # Items 1 and 2 are already processed
        )

        mock_config = Config(radarr=RadarrConfig(url="http://localhost:7878", api_key="key"))
        mock_state_manager = MagicMock(spec=StateManager)
        mock_state_manager.get_batch_progress.return_value = existing_progress
        mock_state_manager.get_stale_unavailable_items.return_value = []

        # Create mock movies
        mock_movies = []
        for i in range(1, 4):
            m = MagicMock()
            m.id = i
            m.title = f"Movie {i}"
            mock_movies.append(m)

        with (
            patch("filtarr.cli.Config.load", return_value=mock_config),
            patch("filtarr.cli.get_state_manager", return_value=mock_state_manager),
            patch("filtarr.cli._fetch_movies_to_check") as mock_fetch,
            patch("filtarr.cli.get_checker") as mock_get_checker,
        ):
            mock_fetch.return_value = (mock_movies, set())
            mock_checker = AsyncMock()
            mock_checker.check_movie.return_value = SearchResult(
                item_id=3, item_type="movie", has_match=True
            )
            mock_get_checker.return_value = mock_checker

            result = runner.invoke(
                app,
                ["check", "batch", "--all-movies", "--format", "simple"],
            )

        # Should have skipped items 1 and 2, only processed item 3
        assert "Summary" in result.output


class TestBatchProgressClearing:
    """Tests for L1034 and L1042: Break from batch and clear progress on completion."""

    def test_batch_clears_progress_on_successful_completion(self) -> None:
        """Should clear batch progress when batch completes successfully."""
        mock_config = Config(radarr=RadarrConfig(url="http://localhost:7878", api_key="key"))
        mock_state_manager = MagicMock(spec=StateManager)
        mock_state_manager.get_batch_progress.return_value = None
        mock_state_manager.get_stale_unavailable_items.return_value = []

        # Create a single mock movie
        mock_movie = MagicMock()
        mock_movie.id = 1
        mock_movie.title = "Test Movie"

        with (
            patch("filtarr.cli.Config.load", return_value=mock_config),
            patch("filtarr.cli.get_state_manager", return_value=mock_state_manager),
            patch("filtarr.cli._fetch_movies_to_check") as mock_fetch,
            patch("filtarr.cli.get_checker") as mock_get_checker,
        ):
            mock_fetch.return_value = ([mock_movie], set())
            mock_checker = AsyncMock()
            mock_checker.check_movie.return_value = SearchResult(
                item_id=1, item_type="movie", has_match=True
            )
            mock_get_checker.return_value = mock_checker

            runner.invoke(
                app,
                ["check", "batch", "--all-movies", "--format", "simple"],
            )

        # Should have cleared batch progress on successful completion
        mock_state_manager.clear_batch_progress.assert_called_once()

    def test_batch_does_not_clear_progress_when_batch_limit_reached(self) -> None:
        """Should NOT clear batch progress when batch limit is reached."""
        mock_config = Config(radarr=RadarrConfig(url="http://localhost:7878", api_key="key"))
        mock_state_manager = MagicMock(spec=StateManager)
        mock_state_manager.get_batch_progress.return_value = None
        mock_state_manager.get_stale_unavailable_items.return_value = []

        # Create a new BatchProgress to be returned by start_batch
        new_progress = BatchProgress(batch_id="new-batch", item_type="movie", total_items=5)
        mock_state_manager.start_batch.return_value = new_progress

        # Create multiple mock movies
        mock_movies = []
        for i in range(1, 6):
            m = MagicMock()
            m.id = i
            m.title = f"Movie {i}"
            mock_movies.append(m)

        with (
            patch("filtarr.cli.Config.load", return_value=mock_config),
            patch("filtarr.cli.get_state_manager", return_value=mock_state_manager),
            patch("filtarr.cli._fetch_movies_to_check") as mock_fetch,
            patch("filtarr.cli.get_checker") as mock_get_checker,
        ):
            mock_fetch.return_value = (mock_movies, set())
            mock_checker = AsyncMock()
            # Return a different result for each movie ID
            mock_checker.check_movie.side_effect = lambda movie_id, **_: SearchResult(
                item_id=movie_id, item_type="movie", has_match=True
            )
            mock_get_checker.return_value = mock_checker

            result = runner.invoke(
                app,
                [
                    "check",
                    "batch",
                    "--all-movies",
                    "--batch-size",
                    "2",  # Limit to 2 items
                    "--format",
                    "simple",
                ],
            )

        # Should NOT have cleared batch progress since batch limit was reached
        mock_state_manager.clear_batch_progress.assert_not_called()
        assert "batch limit" in result.output.lower()


class TestRunBatchChecksNewProgress:
    """Tests for L1004-1006: Creating new batch_progress."""

    @pytest.mark.asyncio
    async def test_run_batch_checks_creates_new_progress_when_none_exists(
        self,
        mock_config: Config,
        real_console: Console,
        mock_state_manager: MagicMock,
    ) -> None:
        """Should create new batch progress when no existing progress."""
        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=real_console,
            error_console=real_console,
        )

        # Create mock movies
        mock_movie = MagicMock()
        mock_movie.id = 1
        mock_movie.title = "Test Movie"

        new_progress = BatchProgress(batch_id="new-batch", item_type="movie", total_items=1)
        mock_state_manager.start_batch.return_value = new_progress

        with (
            patch("filtarr.cli._fetch_movies_to_check", return_value=([mock_movie], set())),
            patch("filtarr.cli._process_single_item", return_value=True),
        ):
            await _run_batch_checks(
                ctx,
                all_movies=True,
                all_series=False,
                skip_tagged=False,
                file_items=[],
                batch_type="movie",
                existing_progress=None,
            )

        # Should have started a new batch
        mock_state_manager.start_batch.assert_called_once()
        # Batch ID should be 8 characters (UUID[:8])
        call_args = mock_state_manager.start_batch.call_args
        assert len(call_args[0][0]) == 8  # batch_id
        assert call_args[0][1] == "movie"  # batch_type
        assert call_args[0][2] == 1  # total_items

    @pytest.mark.asyncio
    async def test_run_batch_checks_uses_existing_progress(
        self,
        mock_config: Config,
        real_console: Console,
        mock_state_manager: MagicMock,
    ) -> None:
        """Should use existing progress when provided."""
        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=real_console,
            error_console=real_console,
        )

        mock_movie = MagicMock()
        mock_movie.id = 1
        mock_movie.title = "Test Movie"

        existing_progress = BatchProgress(
            batch_id="existing-batch",
            item_type="movie",
            total_items=5,
            processed_ids={1},
        )

        with (
            patch("filtarr.cli._fetch_movies_to_check", return_value=([mock_movie], set())),
            patch("filtarr.cli._process_single_item", return_value=True),
        ):
            await _run_batch_checks(
                ctx,
                all_movies=True,
                all_series=False,
                skip_tagged=False,
                file_items=[],
                batch_type="movie",
                existing_progress=existing_progress,
            )

        # Should NOT have started a new batch
        mock_state_manager.start_batch.assert_not_called()


class TestBatchProgressBarShowsBatchSize:
    """Tests for progress bar showing batch size instead of total items."""

    @pytest.mark.asyncio
    async def test_progress_bar_total_is_batch_size_when_set(
        self,
        mock_config: Config,
        real_console: Console,
        mock_state_manager: MagicMock,
    ) -> None:
        """When batch_size is set, progress bar total should be min(batch_size, total_items)."""
        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=5,  # Batch size of 5
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=real_console,
            error_console=real_console,
        )

        # Create 100 mock movies (more than batch size)
        mock_movies = []
        for i in range(1, 101):
            m = MagicMock()
            m.id = i
            m.title = f"Movie {i}"
            mock_movies.append(m)

        new_progress = BatchProgress(batch_id="new-batch", item_type="movie", total_items=100)
        mock_state_manager.start_batch.return_value = new_progress

        # Track how many items are actually processed
        processed_count = 0

        async def mock_process_single(*_args: Any, **_kwargs: Any) -> bool:
            nonlocal processed_count
            processed_count += 1
            # Simulate successful processing
            ctx.processed_this_run += 1
            ctx.results.append(
                SearchResult(item_id=processed_count, item_type="movie", has_match=True)
            )
            # Return False when batch limit reached
            if ctx.batch_size > 0 and ctx.processed_this_run >= ctx.batch_size:
                ctx.batch_limit_reached = True
                return False
            return True

        with (
            patch("filtarr.cli._fetch_movies_to_check", return_value=(mock_movies, set())),
            patch("filtarr.cli._process_single_item", side_effect=mock_process_single),
        ):
            await _run_batch_checks(
                ctx,
                all_movies=True,
                all_series=False,
                skip_tagged=False,
                file_items=[],
                batch_type="movie",
                existing_progress=None,
            )

        # Should have only processed batch_size items (5), not all 100
        assert processed_count == 5
        assert ctx.batch_limit_reached is True
        assert len(ctx.results) == 5

    @pytest.mark.asyncio
    async def test_progress_bar_total_is_all_items_when_batch_size_zero(
        self,
        mock_config: Config,
        real_console: Console,
        mock_state_manager: MagicMock,
    ) -> None:
        """When batch_size is 0 (unlimited), progress bar total should be total items."""
        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=0,  # Unlimited
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=real_console,
            error_console=real_console,
        )

        # Create 10 mock movies
        mock_movies = []
        for i in range(1, 11):
            m = MagicMock()
            m.id = i
            m.title = f"Movie {i}"
            mock_movies.append(m)

        new_progress = BatchProgress(batch_id="new-batch", item_type="movie", total_items=10)
        mock_state_manager.start_batch.return_value = new_progress

        processed_count = 0

        async def mock_process_single(*_args: Any, **_kwargs: Any) -> bool:
            nonlocal processed_count
            processed_count += 1
            ctx.results.append(
                SearchResult(item_id=processed_count, item_type="movie", has_match=True)
            )
            return True

        with (
            patch("filtarr.cli._fetch_movies_to_check", return_value=(mock_movies, set())),
            patch("filtarr.cli._process_single_item", side_effect=mock_process_single),
        ):
            await _run_batch_checks(
                ctx,
                all_movies=True,
                all_series=False,
                skip_tagged=False,
                file_items=[],
                batch_type="movie",
                existing_progress=None,
            )

        # Should have processed all 10 items
        assert processed_count == 10
        assert ctx.batch_limit_reached is False
        assert len(ctx.results) == 10

    @pytest.mark.asyncio
    async def test_progress_bar_total_uses_items_count_when_less_than_batch_size(
        self,
        mock_config: Config,
        real_console: Console,
        mock_state_manager: MagicMock,
    ) -> None:
        """When total items < batch_size, progress bar total should be total items."""
        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=100,  # Batch size larger than items
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=real_console,
            error_console=real_console,
        )

        # Create only 3 mock movies (less than batch size of 100)
        mock_movies = []
        for i in range(1, 4):
            m = MagicMock()
            m.id = i
            m.title = f"Movie {i}"
            mock_movies.append(m)

        new_progress = BatchProgress(batch_id="new-batch", item_type="movie", total_items=3)
        mock_state_manager.start_batch.return_value = new_progress

        processed_count = 0

        async def mock_process_single(*_args: Any, **_kwargs: Any) -> bool:
            nonlocal processed_count
            processed_count += 1
            ctx.results.append(
                SearchResult(item_id=processed_count, item_type="movie", has_match=True)
            )
            return True

        with (
            patch("filtarr.cli._fetch_movies_to_check", return_value=(mock_movies, set())),
            patch("filtarr.cli._process_single_item", side_effect=mock_process_single),
        ):
            await _run_batch_checks(
                ctx,
                all_movies=True,
                all_series=False,
                skip_tagged=False,
                file_items=[],
                batch_type="movie",
                existing_progress=None,
            )

        # Should have processed all 3 items (less than batch_size of 100)
        assert processed_count == 3
        assert ctx.batch_limit_reached is False
        assert len(ctx.results) == 3


# =============================================================================
# TESTS FOR ERROR COLLECTION IN BATCH SUMMARY (Task 7)
# =============================================================================


class TestFormatErrorMessage:
    """Tests for _format_error_message helper function."""

    def test_format_http_status_error(self) -> None:
        """Should format HTTP status errors as 'HTTP XXX'."""
        from filtarr.cli import _format_error_message

        mock_response = MagicMock()
        mock_response.status_code = 404
        mock_request = MagicMock()

        error = httpx.HTTPStatusError("Not Found", request=mock_request, response=mock_response)

        result = _format_error_message(error)
        assert result == "HTTP 404"

    def test_format_connect_error(self) -> None:
        """Should format connection errors as 'Connection failed'."""
        from filtarr.cli import _format_error_message

        error = httpx.ConnectError("Connection refused")

        result = _format_error_message(error)
        assert result == "Connection failed"

    def test_format_timeout_error(self) -> None:
        """Should format timeout errors as 'Request timed out'."""
        from filtarr.cli import _format_error_message

        error = httpx.TimeoutException("Timeout")

        result = _format_error_message(error)
        assert result == "Request timed out"

    def test_format_configuration_error(self) -> None:
        """Should format configuration errors with their message."""
        from filtarr.cli import _format_error_message

        error = ConfigurationError("Radarr not configured")

        result = _format_error_message(error)
        assert result == "Radarr not configured"

    def test_format_generic_error(self) -> None:
        """Should format generic errors with str()."""
        from filtarr.cli import _format_error_message

        error = RuntimeError("Something went wrong")

        result = _format_error_message(error)
        assert result == "Something went wrong"


class TestBatchErrorCollectionInSummary:
    """Tests for error collection and display in batch summary."""

    @pytest.mark.asyncio
    async def test_errors_collected_during_processing(
        self,
        mock_config: Config,
        mock_console: Any,
        mock_error_console: Any,
        mock_state_manager: MagicMock,
    ) -> None:
        """Errors should be collected in the formatter during batch processing."""

        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=mock_console,
            error_console=mock_error_console,
        )

        mock_response = MagicMock()
        mock_response.status_code = 500
        mock_request = MagicMock()

        with patch(
            "filtarr.cli._process_batch_item",
            side_effect=httpx.HTTPStatusError(
                "Internal Server Error", request=mock_request, response=mock_response
            ),
        ):
            await _process_single_item(ctx, "movie", 123, "Test Movie", batch_progress=None)

        # Error should be collected in formatter
        assert len(ctx.formatter.errors) == 1
        assert ctx.formatter.errors[0] == ("Test Movie", "HTTP 500")

    @pytest.mark.asyncio
    async def test_multiple_errors_collected(
        self,
        mock_config: Config,
        mock_console: Any,
        mock_error_console: Any,
        mock_state_manager: MagicMock,
    ) -> None:
        """Multiple errors should be collected in the formatter."""
        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=mock_console,
            error_console=mock_error_console,
        )

        # Simulate different types of errors
        errors = [
            httpx.ConnectError("Connection refused"),
            httpx.TimeoutException("Timeout"),
            ConfigurationError("API key missing"),
        ]

        for i, error in enumerate(errors):
            with patch("filtarr.cli._process_batch_item", side_effect=error):
                await _process_single_item(
                    ctx, "movie", i + 1, f"Movie {i + 1}", batch_progress=None
                )

        # All errors should be collected
        assert len(ctx.formatter.errors) == 3
        assert ctx.formatter.errors[0] == ("Movie 1", "Connection failed")
        assert ctx.formatter.errors[1] == ("Movie 2", "Request timed out")
        assert ctx.formatter.errors[2] == ("Movie 3", "API key missing")

    def test_batch_summary_with_no_errors(
        self,
        mock_config: Config,
        mock_console: Any,
        mock_error_console: Any,
        mock_state_manager: MagicMock,
    ) -> None:
        """Batch summary should not show error section if no errors."""
        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=mock_console,
            error_console=mock_error_console,
        )

        # No errors added

        # Get format_summary output
        summary_lines = ctx.formatter.format_summary()

        # Should be empty if no errors
        assert len(summary_lines) == 0

    @pytest.mark.asyncio
    async def test_error_uses_display_name_for_named_items(
        self,
        mock_config: Config,
        mock_console: Any,
        mock_error_console: Any,
        mock_state_manager: MagicMock,
    ) -> None:
        """Error collection should use the item name for named items."""
        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=mock_console,
            error_console=mock_error_console,
        )

        with patch(
            "filtarr.cli._process_batch_item",
            side_effect=httpx.ConnectError("Connection refused"),
        ):
            await _process_single_item(ctx, "movie", 123, "The Matrix", batch_progress=None)

        # Should use the item name, not "movie:123"
        assert len(ctx.formatter.errors) == 1
        assert ctx.formatter.errors[0][0] == "The Matrix"

    @pytest.mark.asyncio
    async def test_error_uses_type_and_id_for_id_prefix_items(
        self,
        mock_config: Config,
        mock_console: Any,
        mock_error_console: Any,
        mock_state_manager: MagicMock,
    ) -> None:
        """Error collection should use type:id for items with ID: prefix."""
        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=mock_console,
            error_console=mock_error_console,
        )

        with patch(
            "filtarr.cli._process_batch_item",
            side_effect=httpx.TimeoutException("Timeout"),
        ):
            await _process_single_item(ctx, "movie", 456, "ID:456", batch_progress=None)

        # Should use "movie:456" since item_name starts with "ID:"
        assert len(ctx.formatter.errors) == 1
        assert ctx.formatter.errors[0][0] == "movie:456"


class TestBatchContextFormatterInitialization:
    """Tests for BatchContext formatter initialization."""

    def test_batch_context_has_formatter_by_default(
        self,
        mock_config: Config,
        mock_console: Any,
        mock_error_console: Any,
        mock_state_manager: MagicMock,
    ) -> None:
        """BatchContext should have an OutputFormatter by default."""
        from filtarr.output import OutputFormatter

        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=mock_console,
            error_console=mock_error_console,
        )

        # Should have a formatter
        assert ctx.formatter is not None
        assert isinstance(ctx.formatter, OutputFormatter)
        assert ctx.formatter.errors == []
        assert ctx.formatter.warnings == []


class TestPrintBatchSummary:
    """Tests for _print_batch_summary output formatting."""

    def test_print_batch_summary_with_warnings(
        self,
        mock_config: Config,
        mock_console: Any,
        mock_error_console: Any,
        mock_state_manager: MagicMock,
    ) -> None:
        """_print_batch_summary should print warnings in yellow."""
        from filtarr.cli import _print_batch_summary

        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=mock_console,
            error_console=mock_error_console,
        )

        # Add a warning to the formatter
        ctx.formatter.add_warning("Slow request (15s)")

        # Call _print_batch_summary
        with patch("filtarr.cli.console") as patched_console:
            _print_batch_summary(ctx)

            # Check that warnings are printed with yellow formatting
            calls = [str(c) for c in patched_console.print.call_args_list]
            # Should contain a call with yellow and "Warnings"
            assert any("[yellow]" in c and "Warnings" in c for c in calls)

    def test_print_batch_summary_with_errors(
        self,
        mock_config: Config,
        mock_console: Any,
        mock_error_console: Any,
        mock_state_manager: MagicMock,
    ) -> None:
        """_print_batch_summary should print errors in red."""
        from filtarr.cli import _print_batch_summary

        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=mock_console,
            error_console=mock_error_console,
        )

        # Add an error to the formatter
        ctx.formatter.add_error("The Matrix", "HTTP 404")

        # Call _print_batch_summary
        with patch("filtarr.cli.console") as patched_console:
            _print_batch_summary(ctx)

            # Check that errors are printed with red formatting
            calls = [str(c) for c in patched_console.print.call_args_list]
            # Should contain a call with red and "Errors"
            assert any("[red]" in c and "Errors" in c for c in calls)
            # Should contain detail lines with dim formatting
            assert any("[dim]" in c and "The Matrix" in c for c in calls)

    def test_print_batch_summary_with_warnings_and_errors(
        self,
        mock_config: Config,
        mock_console: Any,
        mock_error_console: Any,
        mock_state_manager: MagicMock,
    ) -> None:
        """_print_batch_summary should print both warnings and errors."""
        from filtarr.cli import _print_batch_summary

        ctx = BatchContext(
            config=mock_config,
            state_manager=mock_state_manager,
            search_criteria=SearchCriteria.FOUR_K,
            criteria_str="4k",
            sampling_strategy=SamplingStrategy.RECENT,
            seasons=3,
            apply_tags=False,
            dry_run=False,
            batch_size=0,
            delay=0,
            output_format=OutputFormat.SIMPLE,
            console=mock_console,
            error_console=mock_error_console,
        )

        # Add warnings and errors
        ctx.formatter.add_warning("Slow request")
        ctx.formatter.add_error("Movie 1", "Connection failed")

        # Call _print_batch_summary
        with patch("filtarr.cli.console") as patched_console:
            _print_batch_summary(ctx)

            calls = [str(c) for c in patched_console.print.call_args_list]
            # Both warnings and errors should be printed
            assert any("[yellow]" in c and "Warnings" in c for c in calls)
            assert any("[red]" in c and "Errors" in c for c in calls)



"""Tests for ReleaseTagger."""

from __future__ import annotations

from typing import Any
from unittest.mock import AsyncMock, MagicMock

import httpx
import pytest

from filtarr.config import TagConfig
from filtarr.criteria import SearchCriteria
from filtarr.models.common import Tag
from filtarr.tagger import ReleaseTagger, TagResult


class MockTaggableClient:
    """Mock implementation of TaggableClient for testing."""

    def __init__(
        self, tags: list[Tag] | None = None, item_tags: dict[int, list[int]] | None = None
    ) -> None:
        self._tags = tags or []
        self._item_tags = item_tags or {}
        self.get_tags = AsyncMock(return_value=self._tags)
        self.create_tag = AsyncMock(side_effect=self._create_tag)
        self.add_tag_to_item = AsyncMock()
        self.remove_tag_from_item = AsyncMock()
        self._next_tag_id = 100

    def _create_tag(self, label: str) -> Tag:
        """Create a new tag with auto-incrementing ID."""
        tag = Tag(id=self._next_tag_id, label=label)
        self._next_tag_id += 1
        self._tags.append(tag)
        return tag

    async def get_movie_raw(self, movie_id: int) -> dict[str, Any]:
        """Get raw movie data including tags."""
        return {"id": movie_id, "tags": self._item_tags.get(movie_id, [])}

    async def get_series_raw(self, series_id: int) -> dict[str, Any]:
        """Get raw series data including tags."""
        return {"id": series_id, "tags": self._item_tags.get(series_id, [])}


class TestTagResult:
    """Tests for TagResult dataclass."""

    def test_default_values(self) -> None:
        """Should have sensible defaults."""
        result = TagResult()
        assert result.tag_applied is None
        assert result.tag_removed is None
        assert result.tag_created is False
        assert result.tag_already_present is False
        assert result.tag_error is None
        assert result.dry_run is False

    def test_with_values(self) -> None:
        """Should accept values correctly."""
        result = TagResult(
            tag_applied="4k-available",
            tag_removed="4k-unavailable",
            tag_created=True,
            tag_already_present=False,
            dry_run=False,
        )
        assert result.tag_applied == "4k-available"
        assert result.tag_removed == "4k-unavailable"
        assert result.tag_created is True
        assert result.tag_already_present is False
        assert result.dry_run is False

    def test_tag_already_present_field(self) -> None:
        """Should track when tag was already present."""
        result = TagResult(
            tag_applied="4k-available",
            tag_already_present=True,
        )
        assert result.tag_applied == "4k-available"
        assert result.tag_already_present is True


class TestReleaseTagger:
    """Tests for ReleaseTagger class."""

    def test_init_with_default_config(self) -> None:
        """Should initialize with default TagConfig."""
        tagger = ReleaseTagger(TagConfig())
        assert tagger._config is not None

    def test_init_with_custom_config(self) -> None:
        """Should initialize with custom TagConfig."""
        config = TagConfig(
            pattern_available="{criteria}-yes",
            pattern_unavailable="{criteria}-no",
        )
        tagger = ReleaseTagger(config)
        assert tagger._config.pattern_available == "{criteria}-yes"
        assert tagger._config.pattern_unavailable == "{criteria}-no"

    def test_clear_tag_cache(self) -> None:
        """Should clear the tag cache."""
        tagger = ReleaseTagger(TagConfig())
        # Simulate cached tags
        tagger._tag_cache = {"radarr": [Tag(id=1, label="test")]}
        tagger.clear_tag_cache()
        assert tagger._tag_cache is None


class TestApplyTags:
    """Tests for apply_tags method."""

    @pytest.mark.asyncio
    async def test_apply_tags_with_match(self) -> None:
        """Should apply available tag when has_match is True."""
        client = MockTaggableClient(
            tags=[
                Tag(id=1, label="4k-available"),
                Tag(id=2, label="4k-unavailable"),
            ]
        )
        tagger = ReleaseTagger(TagConfig())

        result = await tagger.apply_tags(
            client=client,
            item_id=123,
            item_type="movie",
            has_match=True,
            criteria=SearchCriteria.FOUR_K,
        )

        assert result.tag_applied == "4k-available"
        assert result.tag_removed == "4k-unavailable"
        assert result.tag_created is False
        assert result.tag_already_present is False
        assert result.tag_error is None
        client.add_tag_to_item.assert_called_once_with(123, 1)
        client.remove_tag_from_item.assert_called_once_with(123, 2)

    @pytest.mark.asyncio
    async def test_apply_tags_without_match(self) -> None:
        """Should apply unavailable tag when has_match is False."""
        client = MockTaggableClient(
            tags=[
                Tag(id=1, label="4k-available"),
                Tag(id=2, label="4k-unavailable"),
            ]
        )
        tagger = ReleaseTagger(TagConfig())

        result = await tagger.apply_tags(
            client=client,
            item_id=123,
            item_type="movie",
            has_match=False,
            criteria=SearchCriteria.FOUR_K,
        )

        assert result.tag_applied == "4k-unavailable"
        assert result.tag_removed == "4k-available"
        assert result.tag_created is False
        assert result.tag_already_present is False
        client.add_tag_to_item.assert_called_once_with(123, 2)
        client.remove_tag_from_item.assert_called_once_with(123, 1)

    @pytest.mark.asyncio
    async def test_apply_tags_creates_missing_tag(self) -> None:
        """Should create tag if it doesn't exist."""
        client = MockTaggableClient(tags=[])  # No existing tags
        tagger = ReleaseTagger(TagConfig())

        result = await tagger.apply_tags(
            client=client,
            item_id=123,
            item_type="movie",
            has_match=True,
            criteria=SearchCriteria.FOUR_K,
        )

        assert result.tag_applied == "4k-available"
        assert result.tag_created is True
        assert result.tag_already_present is False
        client.create_tag.assert_called_once_with("4k-available")
        client.add_tag_to_item.assert_called_once()

    @pytest.mark.asyncio
    async def test_apply_tags_dry_run(self) -> None:
        """Should not apply tags in dry_run mode."""
        client = MockTaggableClient(
            tags=[
                Tag(id=1, label="4k-available"),
                Tag(id=2, label="4k-unavailable"),
            ]
        )
        tagger = ReleaseTagger(TagConfig())

        result = await tagger.apply_tags(
            client=client,
            item_id=123,
            item_type="movie",
            has_match=True,
            criteria=SearchCriteria.FOUR_K,
            dry_run=True,
        )

        assert result.tag_applied == "4k-available"
        assert result.tag_removed == "4k-unavailable"
        assert result.dry_run is True
        client.add_tag_to_item.assert_not_called()
        client.remove_tag_from_item.assert_not_called()
        client.create_tag.assert_not_called()

    @pytest.mark.asyncio
    async def test_apply_tags_with_different_criteria(self) -> None:
        """Should use correct tag names for different criteria."""
        client = MockTaggableClient(
            tags=[
                Tag(id=1, label="imax-available"),
                Tag(id=2, label="imax-unavailable"),
            ]
        )
        tagger = ReleaseTagger(TagConfig())

        result = await tagger.apply_tags(
            client=client,
            item_id=123,
            item_type="movie",
            has_match=True,
            criteria=SearchCriteria.IMAX,
        )

        assert result.tag_applied == "imax-available"
        client.add_tag_to_item.assert_called_once_with(123, 1)

    @pytest.mark.asyncio
    async def test_apply_tags_with_custom_pattern(self) -> None:
        """Should use custom tag pattern from config."""
        config = TagConfig(
            pattern_available="has-{criteria}",
            pattern_unavailable="no-{criteria}",
        )
        client = MockTaggableClient(
            tags=[
                Tag(id=1, label="has-4k"),
            ]
        )
        tagger = ReleaseTagger(config)

        result = await tagger.apply_tags(
            client=client,
            item_id=123,
            item_type="movie",
            has_match=True,
            criteria=SearchCriteria.FOUR_K,
        )

        assert result.tag_applied == "has-4k"
        client.add_tag_to_item.assert_called_once_with(123, 1)

    @pytest.mark.asyncio
    async def test_apply_tags_case_insensitive_matching(self) -> None:
        """Should match tags case-insensitively."""
        client = MockTaggableClient(
            tags=[
                Tag(id=1, label="4K-Available"),  # Different case
            ]
        )
        tagger = ReleaseTagger(TagConfig())

        result = await tagger.apply_tags(
            client=client,
            item_id=123,
            item_type="movie",
            has_match=True,
            criteria=SearchCriteria.FOUR_K,
        )

        assert result.tag_applied == "4k-available"
        assert result.tag_created is False  # Should not create, just use existing
        client.add_tag_to_item.assert_called_once_with(123, 1)

    @pytest.mark.asyncio
    async def test_apply_tags_no_opposite_tag_to_remove(self) -> None:
        """Should handle case when opposite tag doesn't exist."""
        client = MockTaggableClient(
            tags=[
                Tag(id=1, label="4k-available"),
                # 4k-unavailable doesn't exist
            ]
        )
        tagger = ReleaseTagger(TagConfig())

        result = await tagger.apply_tags(
            client=client,
            item_id=123,
            item_type="movie",
            has_match=True,
            criteria=SearchCriteria.FOUR_K,
        )

        assert result.tag_applied == "4k-available"
        assert result.tag_removed is None  # No opposite tag to remove
        client.remove_tag_from_item.assert_not_called()

    @pytest.mark.asyncio
    async def test_apply_tags_uses_cache(self) -> None:
        """Should cache tags and not refetch on subsequent calls."""
        client = MockTaggableClient(
            tags=[
                Tag(id=1, label="4k-available"),
                Tag(id=2, label="4k-unavailable"),
            ]
        )
        tagger = ReleaseTagger(TagConfig())

        # First call
        await tagger.apply_tags(
            client=client,
            item_id=123,
            item_type="movie",
            has_match=True,
            criteria=SearchCriteria.FOUR_K,
        )

        # Second call
        await tagger.apply_tags(
            client=client,
            item_id=456,
            item_type="movie",
            has_match=True,
            criteria=SearchCriteria.FOUR_K,
        )

        # get_tags should only be called once due to caching
        assert client.get_tags.call_count == 1

    @pytest.mark.asyncio
    async def test_apply_tags_tag_already_present_on_item(self) -> None:
        """Should detect when tag is already present on item."""
        # Create client where movie 123 already has tag 1
        client = MockTaggableClient(
            tags=[
                Tag(id=1, label="4k-available"),
                Tag(id=2, label="4k-unavailable"),
            ],
            item_tags={123: [1]},  # Movie 123 already has tag 1
        )
        tagger = ReleaseTagger(TagConfig())

        result = await tagger.apply_tags(
            client=client,
            item_id=123,
            item_type="movie",
            has_match=True,
            criteria=SearchCriteria.FOUR_K,
        )

        assert result.tag_applied == "4k-available"
        assert result.tag_already_present is True
        # Should NOT call add_tag_to_item since tag is already present
        client.add_tag_to_item.assert_not_called()
        # Should still remove the opposite tag if present
        client.remove_tag_from_item.assert_called_once_with(123, 2)

    @pytest.mark.asyncio
    async def test_apply_tags_series_tag_already_present(self) -> None:
        """Should detect when tag is already present on series."""
        # Create client where series 456 already has tag 1
        client = MockTaggableClient(
            tags=[
                Tag(id=1, label="4k-available"),
            ],
            item_tags={456: [1]},  # Series 456 already has tag 1
        )
        tagger = ReleaseTagger(TagConfig())

        result = await tagger.apply_tags(
            client=client,
            item_id=456,
            item_type="series",
            has_match=True,
            criteria=SearchCriteria.FOUR_K,
        )

        assert result.tag_applied == "4k-available"
        assert result.tag_already_present is True
        client.add_tag_to_item.assert_not_called()


class TestApplyTagsErrorHandling:
    """Tests for error handling in apply_tags."""

    @pytest.mark.asyncio
    async def test_http_error_handling(self) -> None:
        """Should handle HTTP errors gracefully."""
        client = MockTaggableClient(
            tags=[
                Tag(id=1, label="4k-available"),
            ]
        )
        # Make add_tag_to_item raise an HTTP error
        response = MagicMock()
        response.status_code = 500
        response.reason_phrase = "Internal Server Error"
        client.add_tag_to_item.side_effect = httpx.HTTPStatusError(
            "Server error", request=MagicMock(), response=response
        )
        tagger = ReleaseTagger(TagConfig())

        result = await tagger.apply_tags(
            client=client,
            item_id=123,
            item_type="movie",
            has_match=True,
            criteria=SearchCriteria.FOUR_K,
        )

        assert result.tag_error is not None
        assert "HTTP 500" in result.tag_error

    @pytest.mark.asyncio
    async def test_network_error_handling(self) -> None:
        """Should handle network errors gracefully."""
        client = MockTaggableClient(
            tags=[
                Tag(id=1, label="4k-available"),
            ]
        )
        client.add_tag_to_item.side_effect = httpx.ConnectError("Connection failed")
        tagger = ReleaseTagger(TagConfig())

        result = await tagger.apply_tags(
            client=client,
            item_id=123,
            item_type="movie",
            has_match=True,
            criteria=SearchCriteria.FOUR_K,
        )

        assert result.tag_error is not None
        assert "Network error" in result.tag_error

    @pytest.mark.asyncio
    async def test_timeout_error_handling(self) -> None:
        """Should handle timeout errors gracefully."""
        client = MockTaggableClient(
            tags=[
                Tag(id=1, label="4k-available"),
            ]
        )
        client.add_tag_to_item.side_effect = httpx.TimeoutException("Request timed out")
        tagger = ReleaseTagger(TagConfig())

        result = await tagger.apply_tags(
            client=client,
            item_id=123,
            item_type="movie",
            has_match=True,
            criteria=SearchCriteria.FOUR_K,
        )

        assert result.tag_error is not None
        assert "Network error" in result.tag_error


class TestApplyMovieTags:
    """Tests for apply_movie_tags convenience method."""

    @pytest.mark.asyncio
    async def test_apply_movie_tags(self) -> None:
        """Should delegate to apply_tags with movie item_type."""
        client = MockTaggableClient(
            tags=[
                Tag(id=1, label="4k-available"),
            ]
        )
        tagger = ReleaseTagger(TagConfig())

        result = await tagger.apply_movie_tags(
            client=client,
            movie_id=123,
            has_match=True,
            criteria=SearchCriteria.FOUR_K,
        )

        assert result.tag_applied == "4k-available"
        client.add_tag_to_item.assert_called_once_with(123, 1)


class TestApplySeriesTags:
    """Tests for apply_series_tags convenience method."""

    @pytest.mark.asyncio
    async def test_apply_series_tags(self) -> None:
        """Should delegate to apply_tags with series item_type."""
        client = MockTaggableClient(
            tags=[
                Tag(id=1, label="hdr-available"),
            ]
        )
        tagger = ReleaseTagger(TagConfig())

        result = await tagger.apply_series_tags(
            client=client,
            series_id=456,
            has_match=True,
            criteria=SearchCriteria.HDR,
        )

        assert result.tag_applied == "hdr-available"
        client.add_tag_to_item.assert_called_once_with(456, 1)


class TestTagCacheIsolation:
    """Tests for tag cache isolation between radarr and sonarr."""

    @pytest.mark.asyncio
    async def test_cache_isolation_movie_and_series(self) -> None:
        """Should maintain separate caches for movies and series."""
        # Separate clients for radarr and sonarr
        radarr_client = MockTaggableClient(
            tags=[
                Tag(id=1, label="4k-available"),
            ]
        )
        sonarr_client = MockTaggableClient(
            tags=[
                Tag(id=10, label="4k-available"),
            ]
        )
        tagger = ReleaseTagger(TagConfig())

        # Apply tags for movie
        await tagger.apply_movie_tags(
            client=radarr_client,
            movie_id=123,
            has_match=True,
        )

        # Apply tags for series (should fetch tags separately)
        await tagger.apply_series_tags(
            client=sonarr_client,
            series_id=456,
            has_match=True,
        )

        # Both clients should have had get_tags called
        radarr_client.get_tags.assert_called_once()
        sonarr_client.get_tags.assert_called_once()

        # Verify correct tag IDs were used
        radarr_client.add_tag_to_item.assert_called_once_with(123, 1)
        sonarr_client.add_tag_to_item.assert_called_once_with(456, 10)
